/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2012  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;

import net.sf.autodao.Finder;
import net.sf.autodao.Limit;
import net.sf.autodao.Named;
import net.sf.autodao.Offset;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.util.Assert;
import org.springframework.util.StringUtils;

/**
 * Helper utils.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
public final class Utils {

  /** Deny instantiation. */
  private Utils() {
  }

  @Nullable
  static Limit getLimit(@NotNull final Annotation[] arr) {
    return getInstance(arr, Limit.class);
  }

  @Nullable
  static Offset getOffset(@NotNull final Annotation[] arr) {
    return getInstance(arr, Offset.class);
  }

  @Nullable
  static Named getNamed(@NotNull final Annotation[] arr) {
    return getInstance(arr, Named.class);
  }

  @Nullable
  private static <T> T getInstance(@NotNull final Annotation[] arr, @NotNull final Class<? extends T> klass) {
    for (final Annotation t : arr)
      if (klass.isInstance(t))
        return klass.cast(t);

    return null;
  }

  static void visitMethodParameters(@NotNull final Method method, @NotNull final ParameterCallback callback) {
    int i = 0;
    final Annotation[][] parameterAnnotations = method.getParameterAnnotations();
    for (final Class<?> type : method.getParameterTypes())
      callback.visit(i, type, parameterAnnotations[i++]);
  }

  interface ParameterCallback {
    void visit(final int index, @NotNull final Class<?> type, @NotNull final Annotation[] annotations);
  }

  public static <T> T visitQuery(@NotNull final Method method, @NotNull final QueryVisitor<T> visitor) {
    final Finder finder = AnnotationUtils.findAnnotation(method, Finder.class);
    if (StringUtils.hasText(finder.query())) {
      return visitor.visitQuery(finder.query());
    } else if (StringUtils.hasText(finder.sqlQuery())) {
      return visitor.visitSQLQuery(finder.sqlQuery());
    } else if (StringUtils.hasText(finder.sqlQueryName())) {
      return visitor.visitSQLNamedQuery(finder.sqlQueryName());
    } else {
      Assert.hasText(finder.queryName());
      return visitor.visitNamedQuery(finder.queryName());
    }
  }

  public interface QueryVisitor<T> {
    T visitNamedQuery(@NotNull final String queryName);

    T visitQuery(@NotNull final String query);

    T visitSQLNamedQuery(@NotNull final String queryName);

    T visitSQLQuery(@NotNull final String query);
  }
}
