/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2012  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl.hibernate;

import java.io.Serializable;

import net.sf.autodao.Dao;
import net.sf.autodao.PersistentEntity;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.jetbrains.annotations.NotNull;
import org.springframework.util.Assert;

/**
 * Hibernate implementation of Entity AutoDAO.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
final class HibernateEntityDao<E extends PersistentEntity<K>, K extends Serializable> extends HibernateDao implements Dao<E, K> {

  /** class of persistent object. */
  @NotNull
  private final Class<E> entityType;

  HibernateEntityDao(@NotNull final Class<E> entityType, @NotNull final SessionFactory sessionFactory) {
    super(sessionFactory);
    this.entityType = entityType;
  }

  /** {@inheritDoc} */
  @Override
  public void create(@NotNull final E o) {
    Session session = getCurrentSession();
    try {
      Assert.isTrue(!session.contains(o), "You're trying to persist already persistent entity: " + o);
      session.persist(o);
    } catch (final HibernateException e) {
      throw convertHibernateAccessException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void delete(@NotNull final E o) {
    Session session = getCurrentSession();
    try {
      Assert.isTrue(session.contains(o), "You're trying to delete transient/detached entity: " + o);
      session.delete(o);
    } catch (final HibernateException e) {
      throw convertHibernateAccessException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public E get(@NotNull final K id) {
    Session session = getCurrentSession();
    try {
      return (E) session.get(entityType, id);
    } catch (final HibernateException e) {
      throw convertHibernateAccessException(e);
    }
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public E merge(@NotNull final E o) {
    Session session = getCurrentSession();
    try {
      return (E) session.merge(o);
    } catch (final HibernateException e) {
      throw convertHibernateAccessException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void refresh(@NotNull final E o) {
    Session session = getCurrentSession();
    try {
      session.refresh(o);
    } catch (final HibernateException e) {
      throw convertHibernateAccessException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void saveOrUpdate(@NotNull final E o) {
    Session session = getCurrentSession();
    try {
      session.saveOrUpdate(o);
    } catch (final HibernateException e) {
      throw convertHibernateAccessException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void update(@NotNull final E o) {
    Session session = getCurrentSession();
    try {
      Assert.isTrue(session.contains(o), "You're trying to update transient/detached entity: " + o);
      session.update(o);
    } catch (final HibernateException e) {
      throw convertHibernateAccessException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public String toString() {
    return "Entity Dao for " + entityType.getName();
  }
}
