/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2012  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl.jpa;

import java.lang.reflect.Method;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Query;

import net.sf.autodao.impl.FinderExecutor;
import net.sf.autodao.impl.Utils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.dao.DataAccessResourceFailureException;
import org.springframework.orm.jpa.EntityManagerFactoryUtils;

/**
 * JPA implementation of AutoDAO.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
class JpaDao implements FinderExecutor {

  @NotNull
  private final EntityManagerFactory entityManagerFactory;

  JpaDao(@NotNull EntityManagerFactory entityManagerFactory) {
    this.entityManagerFactory = entityManagerFactory;
  }

  protected final EntityManager getEntityManager() throws DataAccessResourceFailureException {
    return EntityManagerFactoryUtils.getTransactionalEntityManager(entityManagerFactory);
  }

  @SuppressWarnings({ "unchecked" })
  @NotNull
  @Override
  public <T> List<T> executeListFinder(@NotNull final Method method, @NotNull final Map<String, ?> namedArgs, @Nullable final Integer limit, final @Nullable Integer offset) {
    try {
      return populateParams(createQuery(method, limit, offset), namedArgs).getResultList();
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  @SuppressWarnings({ "unchecked" })
  @NotNull
  @Override
  public <T> List<T> executeListFinder(@NotNull final Method method, @NotNull final Object[] indexedArgs, @Nullable final Integer limit, @Nullable final Integer offset) {
    try {
      return populateParams(createQuery(method, limit, offset), indexedArgs).getResultList();
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  @SuppressWarnings("unchecked")
  @Override
  public <T> T executeSingleFinder(@NotNull final Method method, @NotNull final Map<String, ?> namedArgs) {
    try {
      return (T) populateParams(createQuery(method, null, null), namedArgs).getSingleResult();
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  @SuppressWarnings("unchecked")
  @Override
  public <T> T executeSingleFinder(@NotNull final Method method, @NotNull final Object[] indexedArgs) {
    try {
      return (T) populateParams(createQuery(method, null, null), indexedArgs).getSingleResult();
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  @NotNull
  private Query populateParams(@NotNull final Query q, @NotNull final Map<String, ?> params) {
    for (final Map.Entry<String, ?> o : params.entrySet())
      q.setParameter(o.getKey(), o.getValue());

    return q;
  }

  @NotNull
  private Query populateParams(@NotNull final Query q, @NotNull final Object[] params) {
    int i = 0;

    for (final Object o : params)
      q.setParameter(i++, o);

    return q;
  }

  @NotNull
  private Query createQuery(@NotNull final Method m,
                            @Nullable final Integer limit,
                            @Nullable final Integer offset) {
    final Query q = Utils.visitQuery(m, new Utils.QueryVisitor<Query>() {
      @Override
      public Query visitNamedQuery(@NotNull final String queryName) {
        return getEntityManager().createNamedQuery(queryName);
      }

      @Override
      public Query visitQuery(@NotNull final String query) {
        return getEntityManager().createQuery(query);
      }

      @Override
      public Query visitSQLNamedQuery(@NotNull final String queryName) {
        return getEntityManager().createNamedQuery(queryName);
      }

      @Override
      public Query visitSQLQuery(@NotNull final String query) {
        return getEntityManager().createNativeQuery(query);
      }
    });

    if (limit != null)
      q.setMaxResults(limit);

    if (offset != null)
      q.setFirstResult(offset);

    return q;
  }
}
