/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2012  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl.jpa;

import java.io.Serializable;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;

import net.sf.autodao.Dao;
import net.sf.autodao.PersistentEntity;
import org.jetbrains.annotations.NotNull;
import org.springframework.orm.jpa.EntityManagerFactoryUtils;
import org.springframework.util.Assert;

/**
 * JPA implementation of Entity AutoDAO.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
final class JpaEntityDao<E extends PersistentEntity<K>, K extends Serializable> extends JpaDao implements Dao<E, K> {

  /** class of persistent object. */
  @NotNull
  private final Class<E> entityType;

  JpaEntityDao(@NotNull final Class<E> entityType, @NotNull EntityManagerFactory entityManagerFactory) {
    super(entityManagerFactory);
    this.entityType = entityType;
  }

  /** {@inheritDoc} */
  @Override
  public void create(@NotNull final E o) {
    EntityManager entityManager = getEntityManager();
    try {
      Assert.isTrue(!entityManager.contains(o), "You're trying to persist already persistent entity: " + o);
      entityManager.persist(o);
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void delete(@NotNull final E o) {
    EntityManager entityManager = getEntityManager();
    try {
      Assert.isTrue(entityManager.contains(o), "You're trying to delete transient/detached entity: " + o);
      entityManager.remove(o);
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public E get(@NotNull final K id) {
    final EntityManager entityManager = getEntityManager();
    try {
      return entityManager.find(entityType, id);
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public E merge(@NotNull final E o) {
    final EntityManager entityManager = getEntityManager();
    try {
      return entityManager.merge(o);
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void refresh(@NotNull final E o) {
    final EntityManager entityManager = getEntityManager();
    try {
      entityManager.refresh(o);
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void saveOrUpdate(@NotNull final E o) {
    final EntityManager entityManager = getEntityManager();
    try {
      entityManager.persist(o);
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public void update(@NotNull final E o) {
    EntityManager entityManager = getEntityManager();
    try {
      Assert.isTrue(entityManager.contains(o), "You're trying to update transient/detached entity: " + o);
      entityManager.persist(o);
    } catch (RuntimeException e) {
      throw EntityManagerFactoryUtils.convertJpaAccessExceptionIfPossible(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public String toString() {
    return "Entity Dao for " + entityType.getName();
  }
}
